[comment]: # (This file is generated, DO NOT edit this file!)

# Calculation service for esdl_type {{ esdl_type }}:

{{description}}

## Calculations

{% for calculation in calculations %}
### {{ calculation.name }} 

{{ calculation.description }}
{% if calculation.inputs is defined and calculation.inputs|length > 0 %}
#### Input parameters
|Name            |esdl_type            |data_type            |unit            |description            |
|----------------|---------------------|---------------------|----------------|-----------------------|
{% for param in calculation.inputs %}
|{{ param.name }}|{{ param.esdl_type }}|{{ param.data_type }}|{{ param.unit }}|{{ param.description }}|
{% endfor %}
{% endif %}
{% if calculation.outputs is defined and calculation.outputs|length > 0 %}
#### Output values
|Name             |data_type             |unit             |description             |
|-----------------|----------------------|-----------------|------------------------|
{% for output in calculation.outputs %}
|{{ output.name }}|{{ output.data_type }}|{{ output.unit }}|{{ output.description }}|
{% endfor %}
{% endif %}
{% endfor %}

{% if relevant_links is defined and relevant_links|length > 0 %}
### Relevant links
|Link             |description             |
|-----------------|------------------------|
{% for link in relevant_links %}
|[{{link.name}}]({{link.url}})|{{ link.description }}|
{% endfor %}
{% endif %}